///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//  Copyright  NetworkDLS 2002, All rights reserved
//
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF 
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO 
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A 
// PARTICULAR PURPOSE.
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#ifndef _CSocketServer_ClientProperties_Cpp
#define _CSocketServer_ClientProperties_Cpp
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#define _WIN32_WINNT 0x0500
#define WIN32_LEAN_AND_MEAN

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <WinSock2.H>
#include <WinSock.H>
#include <Windows.H>
#include <Stdio.H>
#include <Stdlib.H>
#include <Time.H>

#include "CSocketServer.H"

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	Returns true if the client is in "packet mode", else returns false.
*/
bool CSocketServer::PacketMode(int iClient)
{
	return this->CCI[iClient].bPacketMode;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	Enables / Disables "packet mode" for a connected client.
	
	New connections will inherit the "Packet Mode" value for the socket server.
*/
void CSocketServer::PacketMode(int iClient, bool bUsePacketing)
{
	this->CCI[iClient].bPacketMode = bUsePacketing;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

HANDLE CSocketServer::ClientEventHandle(int iClient)
{
	return this->CCI[iClient].hEvent;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	Returns a pointer to used defined per client data.
*/
VOID *CSocketServer::UserData(int iClient)
{
	return this->CCI[iClient].pUserData;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	Sets the pointer to used defined per client data.
*/
void CSocketServer::UserData(int iClient, VOID *pData)
{
	this->CCI[iClient].pUserData = pData;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	This function reserves a client slots.

	Reserved slots will not be made availabe to the free pool
		even after client disconnection therfore all client variables are protected.
*/
void CSocketServer::Reserved(int iClient, bool bReserve)
{
	this->CCI[iClient].bReserved = bReserve;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	Returns whether a client slot is reserved or not. See Reserved() overloads for more info.
*/
bool CSocketServer::Reserved(int iClient)
{
	return this->CCI[iClient].bReserved;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	This function is used to get a clients peer Port from the iClientID.
*/
int CSocketServer::PeerPort(int iClientID)
{
	return this->CCI[iClientID].iPeerPort;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	This function is used to get a clients peer IP address from the iClientID.
*/
char *CSocketServer::PeerAddress(int iClientID)
{
	return this->CCI[iClientID].sPeerAddress;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	This function is used to get a clients peer IP address from the iClientID.
*/
int CSocketServer::PeerAddress(int iClientID, char *sOutAddress, int iMaxSz)
{
	return (int) strcpy_s(sOutAddress, iMaxSz, this->CCI[iClientID].sPeerAddress);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	This function is used to get a clients peer ID from the iClientID.
*/
int CSocketServer::PeerID(int iClientID)
{
	return this->CCI[iClientID].iPeerID;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	This function returns TRUE if the client has data waiting to be sent.
*/
bool CSocketServer::IsSendDataWaiting(int iClient)
{
	// Is data waiting to be received
	return (this->CCI[iClient].SendChunk.iLength > 0);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	This function returns TRUE if the client has data in there receive buffer.

	NOTE: A TRUE return value does not mean that the data is a full packet,
	therfore a call to GetClientData or GetNextClientData may not return data.
*/
bool CSocketServer::IsRecvDataWaiting(int iClient)
{
	// Is data waiting to be sent?
	return (this->CCI[iClient].RecvChunk.iLength > 0);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	Sets the specified clients socket to be disconnected. Returns TRUE on success. 
*/
bool CSocketServer::Disconnect(int iClient)
{
	this->Lock();

	if(this->IsConnected(iClient))
	{
		this->CCI[iClient].bDisconnect = true;
		this->Unlock();
		return true;
	}
	else{
		this->Unlock();
		return false;
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	Returns TRUE of the specified client is connected and not pending disconnection. 
*/
bool CSocketServer::IsConnected(int iClient)
{
	this->Lock();
	bool bResult = (this->CCI[iClient].bReserved ||
		(this->CCI[iClient].bConnected && !this->CCI[iClient].bDisconnect));
	this->Unlock();
	return bResult;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	This function will return TRUE when the client has a empty send buffer.
	Returns FALSE if the client has been disconnected or is pending a disconnect.

	Important:
		If you are working within the ClientThreadProc() and wish to send data to a connected peer other than
			the one identified by the "iClientID" passed into the thread proc, then you must pass the ID of
			the client that is sending the message as well as the ID of the client that is to receive the
			message. This is because both connections must be monitored for disconnection in order to
			prevent deadlocks.
*/
bool CSocketServer::WaitOnDataToBeSent(int iSourceClientID, int iTargetClientID)
{
	bool bRetVal = false;

	while(this->IsConnected(iSourceClientID) && this->IsConnected(iTargetClientID)
		&& (bRetVal = this->IsSendDataWaiting(iTargetClientID)))
	{
		Sleep(1);
	}

	return !bRetVal;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	This function will return TRUE when the client has a empty send buffer.
	Returns FALSE if the client has been disconnected or is pending a disconnect.
*/
bool CSocketServer::WaitOnDataToBeSent(int iClient)
{
	bool bRetVal = false;

	while(this->IsConnected(iClient) && (bRetVal = this->IsSendDataWaiting(iClient)))
	{
		Sleep(1);
	}

	return !bRetVal;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	This function will return TRUE when the client has data in there receive buffer.
	Returns FALSE if the client has been disconnected or is pending a disconnect.
	
	NOTE: A TRUE return value does not mean that the data is a full packet,
	therfore a call to GetClientData or GetNextClientData may not return data.
*/
bool CSocketServer::WaitOnDataToBeRecvd(int iClient)
{
	bool bRetVal = false;

	while(this->IsConnected(iClient) && !(bRetVal = this->IsRecvDataWaiting(iClient)))
	{
		Sleep(1);
	}

	return bRetVal;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	This function will return TRUE when the receive buffer is empty.
*/
bool CSocketServer::WaitForEmptyRecvBuffer(int iClient)
{
	bool bRetVal = false;

	while(this->IsConnected(iClient) && (bRetVal = this->IsRecvDataWaiting(iClient)))
	{
		Sleep(1);
	}

	return !bRetVal;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	This function will return TRUE when the client has data in there receive buffer
	and the data was a full packet.

	Returns FALSE if the client has been disconnected, is pending a disconnect or the
	data received is not a full packet.
*/

bool CSocketServer::GetNextClientData(int iClient, LPBASICHUNK pChunk)
{
	while(this->IsConnected(iClient))
	{
		if(this->GetClientData(iClient, pChunk))
		{
			return true;
		}
		else{
			Sleep(1);
		}
	}

	return false;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	This function is used to retrieve a clients data from there receive buffer.

	This function returns TRUE if the clients has data to be received,
	the CriticalSection is open, the data is a full (and valid) packet.
	Otherwise the function returns FALSE.
*/
bool CSocketServer::GetClientData(int iClient, LPBASICHUNK pChunk)
{
	pChunk->iLength = 0;

	if(this->IsRecvDataWaiting(iClient))
	{
		if(!this->TryLock())
		{
			return false;
		}

		int iPacketResult = SSPACKET_UNDEFINED;

		if( (iPacketResult = this->p_DePacketize(iClient, pChunk)) == SSPACKET_PARTIAL_DATA)
		{
			//We don't currently have a full packet, no big deal... it's on its way.
			this->Unlock();
			return false;
		}

		if(iPacketResult <= 0)
		{
			this->CCI[iClient].bDisconnect = true;
			this->Unlock();
			return this->p_ThrowError(1, iClient);
		}

		pChunk->sBuffer[pChunk->iLength] = '\0'; //The class allocates one byte ahead, so this is safe.

		this->Unlock();
		return (iPacketResult == SSPACKET_GOOD_PACKET);
	}
	return false;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/*
	This function is used to set a clients send buffer.

	Returns TRUE if the data was set successfully.
	Returns FALSE if the CriticalSection was not free,
	the client has data waiting to be sent, or if the client
	has been disconnected or is pending a disconnect.
*/
bool CSocketServer::SetSendData(int iClient, const void *sData, int iSize)
{
	//Do we alread have data ready to be sent?
	if(!this->IsSendDataWaiting(iClient))
	{
		if(!this->TryLock())
		{
			return false;
		}

		//Is the client is connected and not ready to be disconnected?
		if(this->IsConnected(iClient))
		{
			//Is our clients send buffer large enough to handle the outgoing data?
			if(iSize >= this->CCI[iClient].SendChunk.iAlloc)
			{
				//Make our send buffer large enough to handle the outgoing data.
				if(!this->ReAlloc(&this->CCI[iClient].SendChunk, iSize + 1))
				{
					this->CCI[iClient].bDisconnect = true;
					this->Unlock();
					return false;
				}
			}

			//Copy the outgoing data to our send buffer.
			memcpy_s(this->CCI[iClient].SendChunk.sBuffer,
				this->CCI[iClient].SendChunk.iAlloc, sData, iSize);
			this->CCI[iClient].SendChunk.iLength = iSize;

			this->CCI[iClient].SendChunk.sBuffer[this->CCI[iClient].SendChunk.iLength] = '\0';
		}
		else{
			this->Unlock();
			return false;
		}

		this->Unlock();
		return true;
	}

	return false;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	This function is used to set a clients send buffer.

	Returns TRUE if the data was set successfully.
	Returns FALSE if the CriticalSection was not free,
	the client has data waiting to be sent, or if the client
	has been disconnected or is pending a disconnect.
*/
bool CSocketServer::SetSendData(int iClient, const char *sData)
{
	return this->SetSendData(iClient, sData, strlen(sData));
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	See: [SetSendData(int iClient, const void *sData, int iSize)] overload.
*/
bool CSocketServer::SetSendData(int iClient, LPBASICHUNK pChunk)
{
	return this->SetSendData(iClient, pChunk->sBuffer, pChunk->iLength);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	This function is used to set a clients send buffer.

	Returns TRUE if the data was set successfully.
	Returns FALSE if the client has been disconnected or is pending a disconnect.

	Important:
		If you are working within the ClientThreadProc() and wish to send data to a connected peer other than
			the one identified by the "iClientID" passed into the thread proc, then you must pass the ID of
			the client that is sending the message as well as the ID of the client that is to receive the
			message. This is because both connections must be monitored for disconnection in order to
			prevent deadlocks.
*/
bool CSocketServer::SetNextSendData(int iSourceClientID, int iTargetClientID, const void *sData, int iSize)
{
	while(this->IsConnected(iSourceClientID) && this->IsConnected(iTargetClientID))
	{
		if(this->SetSendData(iTargetClientID, sData, iSize))
		{
			return true;
		}
		Sleep(1);
	}

	return false;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	This function is used to set a clients send buffer.

	Returns TRUE if the data was set successfully.
	Returns FALSE if the client has been disconnected or is pending a disconnect.
*/
bool CSocketServer::SetNextSendData(int iClient, const void *sData, int iSize)
{
	while(this->IsConnected(iClient))
	{
		if(this->SetSendData(iClient, sData, iSize))
		{
			return true;
		}
		Sleep(1);
	}

	return false;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	This function is used to set a clients send buffer.

	Returns TRUE if the data was set successfully.
	Returns FALSE if the client has been disconnected or is pending a disconnect.

	Important:
		If you are working within the ClientThreadProc() and wish to send data to a connected peer other than
			the one identified by the "iClientID" passed into the thread proc, then you must pass the ID of
			the client that is sending the message as well as the ID of the client that is to receive the
			message. This is because both connections must be monitored for disconnection in order to
			prevent deadlocks.

*/
bool CSocketServer::SetNextSendData(int iSourceClientID, int iTargetClientID, const char *sData)
{
	return this->SetNextSendData(iSourceClientID, iTargetClientID, sData, strlen(sData));
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	This function is used to set a clients send buffer.

	Returns TRUE if the data was set successfully.
	Returns FALSE if the client has been disconnected or is pending a disconnect.

*/
bool CSocketServer::SetNextSendData(int iClient, const char *sData)
{
	return this->SetNextSendData(iClient, sData, strlen(sData));
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	See: [SetNextSendData(int iClient, const char *sData)] overload
*/
bool CSocketServer::SetNextSendData(int iClient, LPBASICHUNK pChunk)
{
	return this->SetNextSendData(iClient, pChunk->sBuffer, pChunk->iLength);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	See: [SetNextSendData(int iSourceClientID, int iTargetClientID, const char *sData)] overload.
*/
bool CSocketServer::SetNextSendData(int iSourceClientID, int iTargetClientID, LPBASICHUNK pChunk)
{
	return this->SetNextSendData(iSourceClientID, iTargetClientID, pChunk->sBuffer, pChunk->iLength);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	This function can be used to poll for connected clients.
	Pass in a pointer to an integer, and it will pass out the client
		slot number of a qualified (connected and not pending disconnection) connected client.
*/
bool CSocketServer::GetNextClient(int *iClient)
{
	static int iLastClient = 0;
	int iStartingPoint = iLastClient;

	if(iLastClient < 0 || iLastClient > this->p_iMaxClients)
	{
		iLastClient = 0;
	}

	while(this->IsActive())
	{
		//Loop from where we left off last time, to the MAX count of clients.
		for(;iLastClient < this->p_iMaxClients; iLastClient++)
		{
			if(this->IsConnected(iLastClient))
			{
				*iClient = iLastClient++;
				return true;
			}
		}
		iLastClient = 0;

		//Loop from slot 0 to where the top loop started, this makes
		//	sure we will never Sleep() while clients are connected.
		for(;iLastClient < iStartingPoint; iLastClient++)
		{
			if(this->IsConnected(iLastClient))
			{
				*iClient = iLastClient++;
				return true;
			}
		}
		iStartingPoint = 0;

		Sleep(1);
	}

	return false;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#endif
